Attribute VB_Name = "vbcMain"
Option Explicit
Option Compare Text

'http://www.xtremevbtalk.com/showthread.php?t=282796
'=========================================================
' mathimagics@yahoo.co.uk
'=========================================================
' MVBLC Link Control Tool:   Module "vbcMain"
'=======================================================
'
' Mathimagics Visual Basic Link Control tool
'
'      Jim White,  July 2006
'      Canberra, Australia
'
'=========================================================================
'
'  A simple tool for customising the linkage of VB6 ActiveX DLL's.
'
'  This is a normal standalone EXE project. Compile it as "Link.exe".
'
'  The real VB6 linker is also called Link.exe, and it is typically
'  located in the same folder as VB6.EXE (in the VB98 folder of
'  your Microsoft Visual Studio installation directory).
'
'      1. Change the real linker's name to "vbLink.exe" (make a
'         backup copy just in case!) in the VB98 folder
'
'      2. Put a copy this MVBLC version of Link.exe into the same
'         folder
'
'  MVBLC will only intervene in the normal linkage procedure if we
'  are making a DLL, and there is a VBC file (a link control command
'  file) present in the same directory as the DLL, and  with the same
'  name as the DLL.  In all other cases MVBLC operates transparently,
'  it just passes the VB6-specified link command over to the real
'  linker.
'
'  A VBC link control file is simply a text file with suffix ".vbc".
'  For example, if you are compiling to c:\fred\vbtest.dll, then MVBLC
'  looks for "c:\fred\vbtest.vbc"
'
'  MVBLC supports the following link control commands:
'
'      EXPORT, ENTRY, TIDY, STATUS
'
'  The Command names are not case-sensitive, but remember that all
'  LINK symbols ARE case-sensitive.  So the the names of modules and
'  functions must match exactly with those used in the project.
'
'  -------------
'  LINK COMMANDS
'  -------------
'
'    EXPORT <module name> <function name list>
'
'       e.g.  Export Module1 Function1 Function2
'             Export Module2 myTest1
'             Export Module2 myTest2
'
'       The nominated functions will be exported. Function list
'           members be in form "Name1 Alias Name2", allowing a
'           function to be exported with 2 ids (handy for C
'           linking, e.g. Export Mod1 vbFunc alias vbFunc@12)
'
'       NOTE: <module name> denotes the Name Property of the
'          corresponding module, NOT its file name!
'
'    ENTRY <module name> <function name>
'
'       The function referenced is exported, and it is marked as the
'       DLL's entrypoint (DllMain) function.
'
'
'  ----------------------
'  MISCELLANEOUS COMMANDS
'  ----------------------
'
'    TIDY
'       VB6 DLL linking produces EXP, LIB and DEF files, which are not
'       usually needed. Include this command in the VBC file and we
'       will remove them after the DLL has been linked.
'
'    STATUS
'       Including this command tells the link tool to display the export
'       table of the new DLL after linking has been completed.
'
'  -----------
'  LINK ERRORS
'  -----------
'
'    If you specify an invalid module or function name, the (real) link
'    process will fail. This program pipes the link log to a temporary
'    file, so if the link fails, it can extract and display the error
'    messages from the real linker.
'
'    If the link fails, a message will also be displayed by the VB6 IDE,
'    with the error message "DLL Load Failed".  After the link stage
'    completes, the IDE attempts to load the DLL it just built, and if
'    the link has failed that DLL will not be available. You should
'    find and fix any errors in the .vbc file and try MAKE again.
'
'========================================================================

'Const VB6FOLDER = "C:\Program Files\DevStudio\VB"
Const VB6FOLDER = "C:\Program Files\MicroSoft Visual Studio\VB98"

Public EXEFILE  As String  ' full pathname of exe/dll file being
Public EXENAME  As String  ' name of exe/dll being built

Dim vbCommand   As String  ' original LINK command line passed in by VB6 IDE
                          
Dim Options()   As String  ' Ccommand line tokens
Dim ObjList()   As String  ' list of project OBJ's being linked

Dim EXEPATH     As String  ' Folder containing OBJ files
Dim xList       As String  ' Export request list

Dim F           As Integer ' file unit

Dim ShowStatus  As Boolean ' flag for STATUS command
Dim TidyFlag    As Boolean ' flag for TIDY   command
Dim NormalLink  As Boolean ' did we modify the link in any way?
Dim eMsg        As String  ' link error message

Sub Main()
   
   NormalLink = True
   vbCommand = Command()   ' make a copy of the command line
   
   '
   ' check command line for diagnostic options
   '
   If InStr(vbCommand, "/STATUS:") Then
      frmLinkInfo.ShowStatus vbCommand
      Exit Sub
      End If
      
   If InStr(vbCommand, "/ERROR:") Then
      frmLinkInfo.ShowError vbCommand
      Exit Sub
      End If
      
   '
   ' If we are making a DLL, look for a .vbc file with the same name
   '
   If InStr(vbCommand, "/DLL") Then LoadVBC
   '
   ' If there are no link customisations, just pass the command
   '   directly to the real VB6 linker
   '
   If NormalLink Then
      Execute "VBLINK " & vbCommand, 1
   Else
      RunCustomlink
      End If
   End Sub

Sub RunCustomlink()
   '
   ' Run the real linker via a batch file, so we can check the results
   '
   F = FreeFile
   Open "c:\vbLink.bat" For Output As #F
   Print #F, Left$(VB6FOLDER, 2) ' ensure project drive selected  '<--added from blog post
   Print #F, "cd """ & VB6FOLDER & """"
   Print #F, "VBLINK " & vbCommand & " 1> c:\vbLink.log"
   Print #F, "del c:\vbLink.bat"   ' make the BAT file tidy up
   Close #F
   Execute "c:\vbLink.bat", 1
   If vbLinkError Then Exit Sub
   If TidyFlag Then Call vbTidy
   If ShowStatus Then DisplayLinkStatus
   End Sub
   
Private Sub LoadVBC()
   
   Dim xFile   As String       ' link control file (dllname.vbc)
   Dim xName   As String       ' dll export  (.DEF) filename
   Dim xKey    As String       ' control file keyword
   Dim mName   As String       ' Module name
   Dim pName() As String       ' proc names to export
   Dim dName   As String       ' temp variable for decorated proc name
   Dim EntryFlag As Boolean    ' true if we find an ENTRY command
   Dim j As Long, k As Long
   
   Options = Split(vbCommand, "/")
   '
   ' We now have:
   '   Options(0)= the LINK command + link object list
   '           1 = the /ENTRY switch
   '           2 = the /OUT   switch
   '           3, 4, etc   other switches /BASE, /SUBSYSTEM, /VERSION, /OPT etc
   '
   ' 1) get the EXEpath and EXEname from the /OUT switch.  It's always the
   '    2nd switch but we search for it just to be certain!
   '
   For k = 1 To UBound(Options)
      If Left$(Options(k), 4) = "OUT:" Then
         EXEFILE = Mid$(Options(k), 5)
         EXEFILE = Trim$(Replace(EXEFILE, """", ""))
         Exit For
         End If
      Next
   
   If EXEFILE = "" Then Exit Sub        ' highly unlikely, but ....
   
   j = InStrRev(EXEFILE, "\")
   EXEPATH = Left$(EXEFILE, j - 1)
   EXENAME = Mid$(EXEFILE, j + 1)
   EXENAME = Left$(EXENAME, Len(EXENAME) - 4)
   
   '
   ' 2) look for the link control file in the EXEPATH folder
   '       then in the project foilder
   '
   xFile = EXEPATH & "\" & EXENAME & ".vbc"
   If Dir$(xFile) = "" Then
      xFile = CurDir & "\" & EXENAME & ".vbc"
      If Dir$(xFile) = "" Then Exit Sub   ' no control file, normal link
      End If
      
   F = FreeFile
   Open xFile For Input As #F
   
   Do Until EOF(F)                        ' ".vbc" Export Control File
      Line Input #F, xKey
      j = InStr(xKey, ";"): If j Then xKey = Left$(xKey, j - 1)
      j = InStr(xKey, "'"): If j Then xKey = Left$(xKey, j - 1)
      
      '
      '  Commands:
      '    Export <moduleid> <function1> <function2> .....
      '    Entry  <moduleid> <function>
      '
      '    Tidy
      '
      ' 3. remove extraneous whitespace and split into tokens
      '
      xKey = Trim(xKey)
      If xKey = "" Then GoTo NextLine
      While InStr(xKey, "  ")
         xKey = Replace(xKey, "  ", " ")
         Wend
      pName = Split(xKey, " ")
      
      Select Case pName(0)
         
         Case "Status":    ShowStatus = True
         
         Case "Tidy":      TidyFlag = True
         
         Case "Export"                     ' EXPORT <module> <proclist>
            If UBound(pName) > 1 Then
               mName = pName(1)            ' module name
               For j = 2 To UBound(pName)
                  dName = "?" & pName(j) & "@" & mName & "@@AAGXXZ" ' decorated name
                  xList = xList & "," & pName(j) & " = " & dName
                  Next
               End If
               
         Case "Entry"                      ' ENTRY <module> <procname>
            If UBound(pName) = 2 Then
               mName = pName(1)            ' module name
               dName = "?" & pName(2) & "@" & mName & "@@AAGXXZ" ' decorated name
               xList = xList & "," & pName(2) & " = " & dName    ' we export it (for convenience only)
               For k = 1 To UBound(Options)
                  If Left$(Options(k), 6) = "ENTRY:" Then
                     Options(k) = "ENTRY:" & pName(2) & " "      ' and adjust the linker
                     vbCommand = " " & Join(Options, "/")        ' \ENTRY switch
                     EntryFlag = True
                     Exit For
                     End If
                  Next
               End If
         
         Case "AddDef"                     ' AddDEF <aliasname> = <name>
            If pName(2) = "=" Then         '   we need to retrieve decorated form of <name>
               dName = pName(3)
               j = InStr(xList, "," & pName(3) & " = ")
               If j Then                   '   if it is in our list
                  dName = Mid(xList, j + Len(pName(3)) + 4)
                  k = InStr(dName, ",")
                  If k Then dName = Left(dName, k - 1)
                  End If
               xList = xList & "," & pName(1) & " = " & dName
               End If
         End Select
NextLine:
      Loop
   Close #F
   If xList = "" Then Exit Sub
   
   '
   ' For a custom link, build a DEF file and add the /DEF switch
   '     to the link command line
   '
   NormalLink = False
   If EntryFlag Then xList = xList & ",__vbaS"
   pName = Split(xList, ",")
   xName = EXEPATH & "\" & EXENAME & ".def"
   
   F = FreeFile
   Open xName For Output As #F
   Print #F, "LIBRARY "; EXENAME
   Print #F, "EXPORTS"
   For j = 1 To UBound(pName)
      Print #F, "   " & pName(j)
      Next
   Close #F
   vbCommand = vbCommand & " /DEF:""" & xName & """"
   End Sub

Public Sub vbTidy()
   '
   '  Run a little batch file to tidy up
   '  after a VB6 dll build
   '
   F = FreeFile
   Open "c:\vbTidy.bat" For Output As #F
   Print #F, Left$(EXEPATH, 2)               ' assert drive in case diferent
   Print #F, "cd """ & EXEPATH & """"
   Print #F, "if exist " & EXENAME & ".exp del " & EXENAME & ".exp"
   Print #F, "if exist " & EXENAME & ".lib del " & EXENAME & ".lib"
   Print #F, "if exist " & EXENAME & ".def del " & EXENAME & ".def"
   Print #F, "del c:\vbTidy.bat"             ' remove this bat file
   Close #F
   Execute "c:\vbTidy.bat", 0                ' doesn't need to be modal
   End Sub


Function vbLinkError() As Boolean
   
   Dim logentry As String, temp As String
   Dim i As Integer, j As Integer
   
   If Dir$(EXEFILE) <> "" Then
      If Dir$("c:\vbLink.log") <> "" Then Kill "c:\vbLink.log"
      FixDLL
      Exit Function ' all is well
      End If
   vbLinkError = True
   Shell VB6FOLDER & "\Link.exe /ERROR:" & EXEFILE, 1
   Exit Function
   
BadSign:
   eMsg = "(can't open link log file - " & Err.Description & ")"
   End Function


Sub DisplayLinkStatus()
   '
   ' following a successful custom DLL link, display the
   ' results (list the Export Table). To avoid interference
   ' with the IDE, we fire up an independent copy of the
   ' link tool, passing it the collected info for display
   '
   Shell VB6FOLDER & "\Link.exe /STATUS:" & EXEFILE, 1
   End Sub
